# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides functionality to run the server."""


import logging
import os
import sys

from .config import Config, config_test, parse_cmd_args, warn_options
from .control import pid_permissions, setgid, setuid
from .daemon import Daemon
from .exceptions import (
    BlackholeRuntimeException,
    ConfigException,
    DaemonException,
)
from .logs import configure_logs
from .supervisor import Supervisor
from .utils import blackhole_config_help


__all__ = ("blackhole_config", "run")
"""Tuple all the things."""


def blackhole_config():
    """
    Print the config help to the console with man-style formatting.

    :raises SystemExit: Exit code :py:obj:`os.EX_OK`.
    """
    logging.basicConfig(format="%(message)s", level=logging.INFO)
    logging.info(blackhole_config_help)
    raise SystemExit(os.EX_OK)


def run():
    """
    Create the asyncio loop and start the server.

    :raises SystemExit: Exit code :py:obj:`os.EX_USAGE` when a configuration
                        error occurs, :py:obj:`os.EX_NOPERM` when a permission
                        error occurs or :py:obj:`os.EX_OK` when the program
                        exits cleanly.
    """
    args = parse_cmd_args(sys.argv[1:])
    configure_logs(args)
    logger = logging.getLogger("blackhole")
    if args.test:
        config_test(args)
    try:
        config = Config(args.config_file).load().test()
        config.args = args
        warn_options(config)
        daemon = Daemon(config.pidfile)
        supervisor = Supervisor()
        pid_permissions()
        setgid()
        setuid()
    except (ConfigException, DaemonException) as err:
        logger.critical(err)
        raise SystemExit(os.EX_USAGE)
    except BlackholeRuntimeException as err:
        logger.critical(err)
        raise SystemExit(os.EX_NOPERM)
    if args.background:
        try:
            daemon.daemonize()
        except DaemonException as err:
            supervisor.close_socks()
            logger.critical(err)
            raise SystemExit(os.EX_NOPERM)
    try:
        supervisor.run()
    except KeyboardInterrupt:
        pass
    raise SystemExit(os.EX_OK)



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Communication protocols used by the worker and child processes."""


import asyncio
import logging

from .config import Config


__all__ = ("StreamReaderProtocol", "PING", "PONG")
"""Tuple all the things."""


logger = logging.getLogger("blackhole.protocols")


PING = b"x01"
"""Protocol message used by the worker and child processes to communicate."""

PONG = b"x02"
"""Protocol message used by the worker and child processes to communicate."""


class StreamReaderProtocol(asyncio.StreamReaderProtocol):
    """The class responsible for handling connections commands."""

    def __init__(self, clients, loop=None):
        """
        Initialise the protocol.

        :param list clients: A list of connected clients.
        :param loop: The event loop to use.
        :type loop: :py:obj:`None` or
                    :py:class:`syncio.unix_events._UnixSelectorEventLoop`
        """
        logger.debug("init")
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        logger.debug("loop")
        super().__init__(
            asyncio.StreamReader(loop=self.loop),
            client_connected_cb=self._client_connected_cb,
            loop=self.loop,
        )
        logger.debug("super")
        self.clients = clients
        self.config = Config()
        logger.debug(self.config)
        # This is not a nice way to do this but, socket.getfqdn silently fails
        # and crashes inbound connections when called after os.fork
        self.fqdn = self.config.mailname

    def flags_from_transport(self):
        """Adapt internal flags for the transport in use."""
        # This has to be done here since passing it as part of init causes
        # flags to become garbled and mixed up. Artifact of loop.create_server
        sock = self.transport.get_extra_info("socket")
        # Ideally this would use transport.get_extra_info('sockname') but that
        # crashes the child process for some weird reason. Getting the socket
        # and interacting directly does not cause a crash, hence...
        sock_name = sock.getsockname()
        flags = self.config.flags_from_listener(sock_name[0], sock_name[1])
        if len(flags.keys()) > 0:
            self._flags = flags
            self._disable_dynamic_switching = True
            logger.debug("Flags enabled, disabling dynamic switching")
            logger.debug(f"Flags for this connection: {self._flags}")

    def _client_connected_cb(self, reader, writer):
        """
        Bind a stream reader and writer to the SMTP Protocol.

        :param asyncio.streams.StreamReader reader: An object for reading
                                                    incoming data.
        :param asyncio.streams.StreamWriter writer: An object for writing
                                                    outgoing data.
        """
        self._reader = reader
        self._writer = writer
        self.clients.append(writer)

    def connection_lost(self, exc):
        """
        Client connection is closed or lost.

        :param exc exc: Exception.
        """
        logger.debug("Peer disconnected")
        super().connection_lost(exc)
        self.connection_closed, self._connection_closed = True, True
        try:
            self.clients.remove(self._writer)
        except ValueError:
            pass

    async def wait(self):
        """
        Wait for data from the client.

        :returns: A line of received data.
        :rtype: :py:obj:`str`

        .. note::

           Also handles client timeouts if they wait too long before sending
           data. -- https://kura.gg/blackhole/configuration.html#timeout
        """
        while not self.connection_closed:
            try:
                line = await asyncio.wait_for(
                    self._reader.readline(),
                    self.config.timeout,
                )
            except asyncio.TimeoutError:
                await self.timeout()
                return None
            return line

    async def close(self):
        """Close the connection from the client."""
        logger.debug("Closing connection")
        if self._writer:
            try:
                self.clients.remove(self._writer)
            except ValueError:
                pass
            self._writer.close()
            await self._writer.drain()
        self._connection_closed = True

    async def push(self, msg):
        """
        Write a response message to the client.

        :param str msg: The message for the SMTP code
        """
        response = f"{msg}\r\n".encode("utf-8")
        logger.debug(f"SEND {response}")
        self._writer.write(response)
        await self._writer.drain()



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides the Smtp protocol wrapper."""


import asyncio
import base64
import inspect
import logging
import random

from .protocols import StreamReaderProtocol
from .utils import message_id


__all__ = ("Smtp",)
"""Tuple all the things."""


logger = logging.getLogger("blackhole.smtp")


class Smtp(StreamReaderProtocol):
    """The class responsible for handling SMTP/SMTPS commands."""

    _bounce_responses = {
        450: "Requested mail action not taken: mailbox unavailable",
        451: "Requested action aborted: local error in processing",
        452: "Requested action not taken: insufficient system storage",
        458: "Unable to queue message",
        521: "Machine does not accept mail",
        550: "Requested action not taken: mailbox unavailable",
        551: "User not local",
        552: "Requested mail action aborted: exceeded storage allocation",
        553: "Requested action not taken: mailbox name not allowed",
        571: "Blocked",
    }
    """The response code and message for each bounce type."""

    _delay = None
    """The delay timer value."""

    _max_delay = 60
    """The maximum delay value in seconds. Cannot be more than 60 seconds."""

    _mode = None
    """The response mode to use."""

    _flags = {}
    """Flags defined in each listen directive."""

    _disable_dynamic_switching = False
    """
    This option disabled dynamic switching functionality.

    Dynamic switching will be disabled when mode= and delay= flags are
    configured.

    https://kura.gg/blackhole/configuration.html#listen
    https://kura.gg/blackhole/configuration.html#tls_listen
    """

    _failed_commands = 0
    """An internal counter of failed commands for a client."""

    def __init__(self, clients, loop=None):
        """
        Initialise the SMTP protocol.

        :param list clients: A list of connected clients.
        :param loop: The event loop to use.
        :type loop: :py:obj:`None` or
                    :py:class:`syncio.unix_events._UnixSelectorEventLoop`

        .. note::

           Loads the configuration, defines the server's FQDN and generates
           an RFC 2822 Message-ID.
        """
        super().__init__(clients, loop)
        self.message_id = message_id(self.fqdn)

    def connection_made(self, transport):
        """
        Tie a connection to blackhole to the SMTP protocol.

        :param asyncio.transports.Transport transport: The transport class.
        """
        super().connection_made(transport)
        logger.debug("Peer connected")
        self.transport = transport
        self.flags_from_transport()
        self.connection_closed = False
        self._handler_coroutine = self.loop.create_task(self._handle_client())

    async def push(self, code, msg):
        """
        Write a response code and message to the client.

        :param int code: SMTP code, i.e. 250.
        :param str msg: The message for the SMTP code
        """
        n_msg = f"{code} {msg}"
        await super().push(n_msg)

    async def _handle_client(self):
        """
        Handle a client connection.

        This method greets the client and then accepts and handles each line
        the client sends, passing off to the correct verb handler.
        """
        await self.greet()
        while not self.connection_closed:
            line = await self.wait()
            if line is None:
                await self.close()
                return
            logger.debug(f"RECV {line}")
            line = line.decode("utf-8").rstrip("\r\n")
            self._line = line
            handler = self.lookup_handler(line)
            if handler:
                await handler()
            else:
                await self.push(502, "5.5.2 Command not recognised")

    def get_auth_members(self):
        """
        Get a list of available AUTH mechanisms.

        :returns: A list of available authentication mechanisms.
        :rtype: :py:obj:`list`
        """
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        cmds = []
        for cmd, _ in members:
            if cmd.startswith("auth_") and cmd != "auth_UNKNOWN":
                cmd = cmd.replace("auth_", "").replace("_", "-")
                cmds.append(cmd)
        return cmds

    def lookup_auth_handler(self, line):
        """
        Look up a handler for the received AUTH mechanism.

        :param str line: A line of data from a client.
        :returns: A callable authentication mechanism.
        :rtype: `blackhole.smtp.Smtp.auth_MECHANISM`

        .. note::

           Using ``pass=`` as part of the auth data will trigger an
           authentication pass, using ``fail=`` will trigger an authentication
           failure.
        """
        parts = line.split(" ")
        if len(parts) < 2:
            return self.auth_UNKNOWN
        mechanism = parts[1].upper()
        if mechanism == "CRAM-MD5":
            return self.auth_CRAM_MD5
        if mechanism not in self.get_auth_members():
            return self.auth_UNKNOWN
        if len(parts) == 3 and mechanism == "PLAIN":
            if "fail=" in line:
                return self._auth_failure
            return self._auth_success
        return getattr(self, f"auth_{mechanism.upper()}", self.auth_UNKNOWN)

    async def auth_UNKNOWN(self):
        """Response to an unknown auth mechanism."""
        await self.push(501, "5.5.4 Syntax: AUTH mechanism")

    async def help_AUTH(self):
        """
        Send help for AUTH mechanisms.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        mechanisms = " ".join(self.get_auth_members())
        await self.push(250, f"Syntax: AUTH {mechanisms}")

    async def auth_LOGIN(self):
        """
        Handle an AUTH LOGIN request.

            >>> AUTH LOGIN
            334 VXNlcm5hbWU6
            >>> pass=letmein
            235 2.7.0 Authentication successful

            >>> AUTH LOGIN
            334 VXNlcm5hbWU6
            >>> fail=letmein
            535 5.7.8 Authentication failed

        .. note::

           Using ``pass=`` as part of the auth data will trigger an
           authentication pass, using ``fail=`` will trigger an authentication
           failure.
        """
        await self.push(334, "VXNlcm5hbWU6")
        line = await self.wait()
        logger.debug(f"RECV {line}")
        if b"fail=" in line.lower():
            await self._auth_failure()
        else:
            await self._auth_success()

    async def auth_CRAM_MD5(self):
        """
        Handle an AUTH CRAM-MD5 request.

            >>> AUTH CRAM-MD5
            334 PDE0NjE5MzA1OTYwMS4yMDQ5LjEyMzI4NTE2...
            >>> pass=letmein
            235 2.7.0 Authentication successful

            >>> AUTH CRAM-MD5
            334 PDE0NjE5MzA1OTYwMS4yMDQ5LjEyMzI4NTE2...
            >>> fail=letmein
            535 5.7.8 Authentication failed

        .. note::

           Using ``pass=`` as part of the auth data will trigger an
           authentication pass, using ``fail=`` will trigger an authentication
           failure.
        """
        emessage_id = base64.b64encode(self.message_id.encode("utf-8"), b"==")
        await self.push(334, emessage_id.decode("utf-8"))
        line = await self.wait()
        logger.debug(f"RECV {line}")
        if b"fail=" in line.lower():
            await self._auth_failure()
        else:
            await self._auth_success()

    async def auth_PLAIN(self):
        """Handle an AUTH PLAIN request.

            >>> AUTH PLAIN
            334
            >>> pass=letmein
            235 2.7.0 Authentication successful

            >>> AUTH PLAIN
            334
            >>> fail=letmein
            535 5.7.8 Authentication failed

            >>> AUTH PLAIN pass=letmein
            235 2.7.0 Authentication successful

            >>> AUTH PLAIN fail=letmein
            535 5.7.8 Authentication failed

        .. note::

           Using ``pass=`` as part of the auth data will trigger an
           authentication pass, using ``fail=`` will trigger an authentication
           failure.
        """
        await self.push(334, " ")
        line = await self.wait()
        logger.debug(f"RECV {line}")
        if b"fail=" in line.lower():
            await self._auth_failure()
        else:
            await self._auth_success()

    async def _auth_success(self):
        """Send an authentication successful response."""
        await self.push(235, "2.7.0 Authentication successful")

    async def _auth_failure(self):
        """Send an authentication failure response."""
        await self.push(535, "5.7.8 Authentication failed")

    async def timeout(self):
        """
        Timeout a client connection.

        Sends the 421 timeout message to the client and closes the connection.

        https://kura.gg/blackhole/configuration.html#timeout
        """
        logger.debug(
            f"Peer timed out, no data received for {self.config.timeout} "
            "seconds",
        )
        await self.push(421, "Timeout")
        await self.close()

    def lookup_handler(self, line):
        """
        Look up the SMTP VERB against a handler.

        :param str line: Look up the command handler to use from the data
                         provided.
        :returns: A callable command handler.
        :rtype: `blackhole.smtp.Smtp.do_VERB`,
                `blackhole.smtp.Smtp.auth_MECHANISM`,
                `blackhole.smtp.Smtp.help_VERB`
        """
        parts = line.split(None, 1)
        if parts:
            if parts[0].lower() == "help":
                return self.lookup_help_handler(parts)
            if parts[0].lower() == "auth":
                return self.lookup_auth_handler(line)
            else:
                return self.lookup_verb_handler(parts[0])
        return self.do_UNKNOWN

    def lookup_help_handler(self, parts):
        """
        Look up a help handler for the SMTP VERB.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help

        :param list parts: A list of command data, split on spaces.
        :returns: A callable help handler.
        :rtype: `blackhole.smtp.Smtp.help_VERB`
        """
        if len(parts) > 1:
            cmd = f"help_{parts[1].upper()}"
        else:
            cmd = "do_HELP"
        return getattr(self, cmd, self.help_UNKNOWN)

    def lookup_verb_handler(self, verb):
        """
        Look up a handler for the SMTP VERB.

        :param str verb:
        :returns: A callable command handler.
        :rtype: `blackhole.smtp.Smtp.do_VERB`
        """
        return getattr(self, f"do_{verb.upper()}", self.do_UNKNOWN)

    async def greet(self):
        """Send a greeting to the client."""
        await self.push(220, f"{self.fqdn} ESMTP")

    def get_help_members(self):
        """
        Get a list of HELP handlers for verbs.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help

        :returns: A list of available help handlers.
        :rtype: :py:obj:`list`
        """
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        cmds = []
        for cmd, _ in members:
            if cmd.startswith("help_") and not cmd == "help_UNKNOWN":
                cmds.append(cmd.replace("help_", ""))
        return cmds

    async def do_HELP(self):
        """
        Send a response to the HELP verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        msg = " ".join(self.get_help_members())
        await self.push(250, f"Supported commands: {msg}")

    async def help_HELO(self):
        """
        Send help for HELO verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: HELO domain.tld")

    async def do_HELO(self):
        """Send response to HELO verb."""
        await self.push(250, "OK")

    async def help_EHLO(self):
        """
        Send help for the EHLO verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: EHLO domain.tld")

    async def do_EHLO(self):
        """Send response to EHLO verb."""
        response = f"250-{self.fqdn}\r\n".encode("utf-8")
        self._writer.write(response)
        logger.debug(f"SENT {response}")
        auth = " ".join(self.get_auth_members())
        responses = (
            "250-HELP",
            "250-PIPELINING",
            f"250-AUTH {auth}",
            f"250-SIZE {self.config.max_message_size}",
            "250-VRFY",
            "250-ETRN",
            "250-ENHANCEDSTATUSCODES",
            "250-8BITMIME",
            "250-SMTPUTF8",
            "250-EXPN",
            "250 DSN",
        )
        for response in responses:
            response = f"{response}\r\n".encode("utf-8")
            logger.debug(f"SENT {response}")
            self._writer.write(response)
        await self._writer.drain()

    async def help_MAIL(self):
        """
        Send help for the MAIL TO verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: MAIL FROM: <address>")

    async def _size_in_mail(self):
        """
        Handle ``SIZE=`` being passed in ``MAIL`` verb.

        Send a 552 response if the size provided is larger than
        max_message_size.
        """
        parts = self._line.lower().split(" ")
        size = None
        for part in parts:
            if part.startswith("size="):
                size = part.split("=")[1]
        if (
            size is not None
            and size.isdigit()
            and int(size) > self.config.max_message_size
        ):
            await self.push(
                552,
                "Message size exceeds fixed maximum message size",
            )
        else:
            await self.push(250, "2.1.0 OK")

    async def do_MAIL(self):
        """
        Send response to MAIL TO verb.

        .. note::

           Checks to see if ``SIZE=`` is passed, pass function off to have it's
           size handled.
        """
        if "size=" in self._line.lower():
            await self._size_in_mail()
        else:
            await self.push(250, "2.1.0 OK")

    async def help_RCPT(self):
        """
        Send response to the RCPT TO verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: RCPT TO: <address>")

    async def do_RCPT(self):
        """Send response to RCPT TO verb."""
        await self.push(250, "2.1.5 OK")

    async def help_DATA(self):
        """
        Send help for the DATA verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: DATA")

    def process_header(self, line):
        """
        Process dynamic switch email headers.

        Reads x-blackhole-delay and x-blackhole-mode headers and re-configures
        on-the-fly how the email is handled based on these headers.

        https://kura.gg/blackhole/dynamic-switches.html

        :param str line: An email header.
        """
        logger.debug(f"HEADER RECV: {line}")
        if self.config.dynamic_switch is False:
            logger.debug("Dynamic switches disabled, ignoring")
            return
        if self._disable_dynamic_switching is True:
            logger.debug("Dynamic switches are disabled by flags option.")
            return
        key, value = line.split(":")
        key, value = key.lower().strip(), value.lower().strip()
        if key == "x-blackhole-delay":
            self.delay = value
        if key == "x-blackhole-mode":
            self.mode = value

    async def response_from_mode(self):
        """
        Send a response based on the configured response mode.

        https://kura.gg/blackhole/dynamic-switches.html#dynamic-delay-switches
        https://kura.gg/blackhole/configuration.html#mode

        Response mode is configured in configuration file and can be overridden
        by email headers, if enabled.
        """
        logger.debug(f"MODE: {self.mode}")
        if self.mode == "bounce":
            key = random.choice(list(self._bounce_responses.keys()))  # nosec
            await self.push(key, self._bounce_responses[key])
        elif self.mode == "random":
            resps = {250: f"2.0.0 OK: queued as {self.message_id}"}
            resps.update(self._bounce_responses)
            key = random.choice(list(resps.keys()))  # nosec
            await self.push(key, resps[key])
        else:
            msg = f"2.0.0 OK: queued as {self.message_id}"
            await self.push(250, msg)

    async def do_DATA(self):
        r"""
        Send response to DATA verb and wait for mail data.

        This method will also implement timeout management and handling after
        receiving the DATA command and no new data is received.

        This method also implements the delay functionality, delaying a
        response after the final '\r\n.\r\n' line. --
        https://kura.gg/blackhole/configuration.html#delay
        https://kura.gg/blackhole/dynamic-switches.html#dynamic-delay-switches

        This method implements restrictions on message sizes. --
        https://kura.gg/blackhole/configuration.html#max-message-size
        """
        await self.push(354, "End data with <CR><LF>.<CR><LF>")
        on_body = False
        msg = []
        while not self.connection_closed:
            line = await self.wait()
            logger.debug(f"RECV {line}")
            msg.append(line)
            if line.lower().startswith(b"x-blackhole") and on_body is False:
                self.process_header(line.decode("utf-8").rstrip("\n"))
            if line == b"\n":
                on_body = True
            if line == b".\r\n":
                break
        if len(b"".join(msg)) > self.config.max_message_size:
            msg = []
            await self.push(
                552,
                "Message size exceeds fixed maximum message size",
            )
            return
        if self.delay:
            logger.debug(f"DELAYING RESPONSE: {self.delay} seconds")
            await asyncio.sleep(self.delay)
        await self.response_from_mode()

    async def do_STARTTLS(self):
        """STARTTLS is not implemented."""
        # It's currently not possible to implement STARTTLS due to lack of
        # support in asyncio. - https://bugs.python.org/review/23749/
        await self.do_NOT_IMPLEMENTED()

    async def help_NOOP(self):
        """
        Send help for the NOOP verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: NOOP")

    async def do_NOOP(self):
        """Send response to the NOOP verb."""
        await self.push(250, "2.0.0 OK")

    async def help_RSET(self):
        """
        Send help for the RSET verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: RSET")

    async def do_RSET(self):
        """
        Send response to the RSET verb.

        A new message id is generated and assigned.
        """
        old_msg_id = self.message_id
        self.message_id = message_id(self.fqdn)
        logger.debug(f"{old_msg_id} is now {self.message_id}")
        await self.push(250, "2.0.0 OK")

    async def help_VRFY(self):
        """
        Send help for the VRFY verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: VRFY <address>")

    async def do_VRFY(self):
        """
        Send response to the VRFY verb.

            >>> VRFY pass=user@domain.tld
            250 2.0.0 <pass=user@domain.tld> OK

            >>> VRFY fail=user@domain.tld
            550 5.7.1 <fail=user@domain.tld> unknown

            >>> VRFY user@domain.tld
            252 2.0.0 Will attempt delivery

        .. note::

           If the request contains 'pass=', the server will respond with code
           250. If the request contains 'fail=', the server will respond with
           code 550. And finally, if neither flag is found, the server will
           respond with code 252.
        """
        _, addr = self._line.split(" ")
        if "pass=" in self._line:
            await self.push(250, f"2.0.0 <{addr}> OK")
        elif "fail=" in self._line:
            await self.push(550, f"5.7.1 <{addr}> unknown")
        else:
            await self.push(252, "2.0.0 Will attempt delivery")

    async def help_EXPN(self):
        """
        Send help for the EXPN verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: EXPN <list1 | list2 | list3 | all>")

    async def _expn_value_to_list(self):
        """
        Look up and return a mailing list or generate one for EXPN all.

        :returns: A list of members for a mailing list.
        :rtype: :py:obj:`list`
        """
        _, expn = self._line.lower().split(" ")
        expn = expn.replace("<", "").replace(">", "")
        lists = {
            "list1": ("Shadow", "Wednesday", "Low-key Liesmith"),
            "list2": (
                "Jim Holden",
                "Naomi Nagata",
                "Alex Kamal",
                "Amos Burton",
            ),
            "list3": (
                "Takeshi Kovacs",
                "Laurens Bancroft",
                "Kristin Ortega",
                "Quellcrist Falconer",
                "Virginia Vidaura",
                "Reileen Kawahara",
            ),
        }
        if expn == "all":
            iterator = []
            for key in lists.keys():
                iterator.extend(lists[key])
            return iterator
        else:
            return lists[expn]

    async def _expn_response(self):
        """
        Generate response for an EXPN query.

        :returns: A list of responses.
        :rtype: :py:obj:`list`
        """
        iterator = await self._expn_value_to_list()
        i, resp = 1, []
        for item in iterator:
            start = "250-"
            if len(iterator) == i:
                start = "250 "
            user = item.lower().replace(" ", ".")
            resp.append(f"{start}{item} <{user}@{self.fqdn}>")
            i += 1
        return resp

    async def do_EXPN(self):
        """
        Handle the EXPN verb.

            >>> EXPN fail=test-list
            550 Not authorised

            >>> EXPN list1
            250-Shadow <shadow@blackhole.io>
            250-Wednesday <wednesday@blackhole.io>
            250 Low-key Liesmith <low-key.liesmith@blackhole.io>

            >>> EXPN list2
            250-Jim Holden <jim.holden@blackhole.io>
            250-Naomi Nagata <naomi.nagata@blackhole.io>
            250-Alex Kamal <alex.kamal@blackhole.io>
            250 Amos Burton <amos.burton@blackhole.io>

            >>> EXPN list3
            250-Takeshi Kovacs <takeshi.kovacs@blackhole.io>
            250-Laurens Bancroft <laurens.bancroft@blackhole.io>
            250-Kristin Ortega <kristin.ortega@blackhole.io>
            250-Quellcrist Falconer <quellcrist.falconer@blackhole.io>
            250-Virginia Vidaura <virginia.vidaura@blackhole.io>
            250 Reileen Kawahara <reileen.kawahara@blackhole.io>

            >>> EXPN all
            250-Shadow <shadow@blackhole.io>
            250-Wednesday <wednesday@blackhole.io>
            250-Low-key Liesmith <low-key.liesmith@blackhole.io>
            250-Takeshi Kovacs <takeshi.kovacs@blackhole.io>
            250-Laurens Bancroft <laurens.bancroft@blackhole.io>
            250-Kristin Ortega <kristin.ortega@blackhole.io>
            250-Quellcrist Falconer <quellcrist.falconer@blackhole.io>
            250-Virginia Vidaura <virginia.vidaura@blackhole.io>
            250-Reileen Kawahara <reileen.kawahara@blackhole.io>
            250-Jim Holden <jim.holden@blackhole.io>
            250-Naomi Nagata <naomi.nagata@blackhole.io>
            250-Alex Kamal <alex.kamal@blackhole.io>
            250 Amos Burton <amos.burton@blackhole.io>

            >>> EXPN list-does-not-exist
            550 Not authorised

        .. note::

           If EXPN contains 'fail=' or does not specify a mailing list the
           command will return a 550 code.
           Valid lists are: `list1`, `list2`, `list3` and `all`.
        """
        if "fail=" in self._line:
            await self.push(550, "Not authorised")
            return
        try:
            _, expn = self._line.lower().split(" ")
            expn = expn.replace("<", "").replace(">", "")
        except ValueError:
            await self.push(550, "Not authorised")
            return
        if expn not in ("list1", "list2", "list3", "all"):
            await self.push(550, "Not authorised")
            return
        for response in await self._expn_response():
            response = f"{response}\r\n".encode("utf-8")
            logger.debug(f"SENT {response}")
            self._writer.write(response)
        await self._writer.drain()

    async def help_ETRN(self):
        """
        Send help for the ETRN verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: ETRN")

    async def do_ETRN(self):
        """Send response to the ETRN verb."""
        await self.push(250, "Queueing started")

    async def help_QUIT(self):
        """
        Send help for the QUIT verb.

        https://kura.gg/blackhole/communicating-with-blackhole.html#help
        """
        await self.push(250, "Syntax: QUIT")

    async def do_QUIT(self):
        """
        Send response to the QUIT verb.

        Closes the client connection.
        """
        await self.push(221, "2.0.0 Goodbye")
        self._handler_coroutine.cancel()
        await self.close()

    async def do_NOT_IMPLEMENTED(self):
        """Send a not implemented response."""
        await self.push(500, "Not implemented")

    async def help_UNKNOWN(self):
        """Send available help verbs when an invalid verb is received."""
        msg = " ".join(self.get_help_members())
        await self.push(501, f"Supported commands: {msg}")

    async def do_UNKNOWN(self):
        """Send response to unknown verb."""
        self._failed_commands += 1
        if self._failed_commands > 9:
            await self.push(502, "5.5.3 Too many unknown commands")
            await self.close()
        else:
            await self.push(502, "5.5.2 Command not recognised")

    @property
    def delay(self):
        """
        Delay after the DATA command completes.

        Value is in seconds, with a maximum value of 60 seconds.

        https://kura.gg/blackhole/configuration.html#delay
        https://kura.gg/blackhole/dynamic-switches.html#dynamic-delay-switches

        :returns: A delay time in seconds. Default: ``None``.
        :rtype: :py:obj:`int` or :py:obj:`None`
        """
        if "delay" in self._flags.keys():
            delay = self._flags["delay"]
            if isinstance(delay, list):
                self._delay_range(delay)
                return self._delay
            return int(delay)
        if self._delay is not None:
            return self._delay
        if self.config.delay is not None:
            return self.config.delay
        return None

    @delay.setter
    def delay(self, values):
        logger.debug("DELAY: Dymanic delay enabled")
        value = values.split(",")
        if len(value) == 2:
            self._delay_range(value)
        elif len(value) == 1:
            self._delay_single(value[0])
        else:
            logger.debug(f"DELAY: Invalid value(s): {values}. Skipping")
            return

    def _delay_range(self, value):
        """
        Generate a delay from a range provided in the email header.

        https://kura.gg/blackhole/configuration.html#delay
        https://kura.gg/blackhole/dynamic-switches.html#dynamic-delay-switches

        :param str value: A list of minimum and maximum values as a string.
                          i.e. (10, 20).

        .. note::

           Converted from a string of a list to a list of integers.
        """
        min_delay, max_delay = value
        min_delay, max_delay = min_delay.strip(), max_delay.strip()
        try:
            min_delay = int(min_delay)
            max_delay = int(max_delay)
        except ValueError:
            logger.debug(
                f"DELAY: Unable to convert {min_delay}, {max_delay} to "
                "integers. Skipping",
            )
            self._delay = None
            return
        if min_delay < 0 or max_delay < 0:
            logger.debug(
                f"DELAY: A value is less than 0: {min_delay}, {max_delay}. "
                "Skipping",
            )
            self._delay = None
            return
        if min_delay > max_delay:
            logger.debug("Min cannot be greater than max")
            self._delay = None
            return
        if max_delay > self._max_delay:
            logger.debug(
                f"DELAY: {max_delay} is higher than {self._max_delay}. "
                f"{self._max_delay} is the hard coded maximum delay for "
                "security.",
            )
            max_delay = self._max_delay
        self._delay = random.randint(min_delay, max_delay)  # nosec
        logger.debug(
            f"DELAY: Set to {self._delay} from range {min_delay}-{max_delay}",
        )
        return

    def _delay_single(self, value):
        """
        Generate a delay from a value provided in an email header.

        https://kura.gg/blackhole/configuration.html#delay
        https://kura.gg/blackhole/dynamic-switches.html#dynamic-delay-switches

        :param str value: Time in seconds as a string.

        .. note:

           Converted from a string to an integer.
        """
        try:
            value = int(value)
        except ValueError:
            logger.debug(
                f"DELAY: Unable to convert {value} to an integer. Skipping",
            )
            self._delay = None
            return
        logger.debug(value)
        if value < 0:
            logger.debug(f"DELAY: {value} is less than 0. Skipping")
            self._delay = None
            return
        if value > self._max_delay:
            logger.debug(
                f"DELAY: {value} is higher than {self._max_delay}. "
                f"{self._max_delay} is the hard coded maximum delay for "
                "security.",
            )
            self._delay = self._max_delay
            return
        logger.debug(f"DELAY: Set to {value}")
        self._delay = value

    @property
    def mode(self):
        """
        How to respond to an email, based on configuration.

        Response is configured in the configuration file or configured from
        email headers, if configured to allow that option.

        https://kura.gg/blackhole/configuration.html#mode
        https://kura.gg/blackhole/dynamic-switches.html#dynamic-mode-switches

        :returns: A response mode.
        :rtype: :py:obj:`str`
        """
        if "mode" in self._flags.keys():
            return self._flags["mode"]
        if self._mode is not None:
            return self._mode
        return self.config.mode

    @mode.setter
    def mode(self, value):
        if value not in ("accept", "bounce", "random"):
            logger.debug(
                f"MODE: {value} is an invalid. Allowed modes: (accept, "
                "bounce, random)",
            )
            self._mode = None
            return
        logger.debug(f"MODE: Dynamic mode enabled. Mode set to {value}")
        self._mode = value



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides daemonisation functionality."""


import atexit
import logging
import os

from .exceptions import DaemonException
from .utils import Singleton


__all__ = ("Daemon",)
"""Tuple all the things."""


logger = logging.getLogger("blackhole.daemon")


class Daemon(metaclass=Singleton):
    """An object for handling daemonisation."""

    def __init__(self, pidfile):
        """
        Create an instance of :class:`Daemon`.

        :param str pidfile: Path to store the pid.

        .. note::

           Registers an :py.func:`atexit.register` signal to delete the pid on
           exit.
        """
        self.pidfile = pidfile
        self.pid = os.getpid()
        atexit.register(self._exit)

    def daemonize(self):
        """Daemonize the process."""
        self.fork()
        os.chdir(os.path.sep)
        os.setsid()
        os.umask(0)
        self.pid = os.getpid()

    def _exit(self, *args, **kwargs):
        """Call on exit using :py:func:`atexit.register` or via a signal."""
        del self.pid

    def fork(self):
        """
        Fork off the process.

        :raises SystemExit: With code :py:obj:`os.EX_OK` when fork is
                            successful.
        :raises DaemonException: When fork is unsuccessful.
        """
        try:
            pid = os.fork()
            if pid > 0:
                os._exit(os.EX_OK)
        except OSError as err:
            raise DaemonException(err.strerror)

    @property
    def pid(self):
        """
        Pid of the process, if it's been daemonised.

        :raises DaemonException: When pid cannot be read from the filesystem.
        :returns: The current pid.
        :rtype: :py:obj:`int` or :py:obj:`None`

        .. note::

           The pid is retrieved from the filestem.
           If the pid does not exist in /proc, the pid is deleted.
        """
        if os.path.exists(self.pidfile):
            try:
                with open(self.pidfile, "r") as pidfile:
                    pid = pidfile.read().strip()
                    if pid != "":
                        return int(pid)
            except IOError as err:
                raise DaemonException(str(err))
        return None

    @pid.setter
    def pid(self, pid):
        """
        Write the pid to the filesystem.

        :param int pid: Process pid.
        :raises DaemonException: When writing to filesystem fails.
        """
        pid = str(pid)
        try:
            with open(self.pidfile, "w+") as pidfile:
                pidfile.write(f"{pid}\n")
        except IOError as err:
            raise DaemonException(str(err))

    @pid.deleter
    def pid(self):
        """Delete the pid from the filesystem."""
        if os.path.exists(self.pidfile):
            os.remove(self.pidfile)



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Exceptions..."""


__all__ = ("ConfigException", "DaemonException", "BlackholeRuntimeException")
"""Tuple all the things."""


class ConfigException(Exception):  # pragma: no cover
    """Configuration exception."""

    pass


class DaemonException(Exception):  # pragma: no cover
    """A daemon exception."""

    pass


class BlackholeRuntimeException(Exception):  # pragma: no cover
    """Blackhole runtime exception."""

    pass



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides functionality to manage child processes from the supervisor."""


import asyncio
import logging
import os
import signal
import time

from . import protocols
from .child import Child
from .control import setgid, setuid
from .streams import StreamProtocol


try:
    import setproctitle
except ImportError:
    setproctitle = None


__all__ = ("Worker",)


logger = logging.getLogger("blackhole.worker")


class Worker:
    """
    A worker.

    Providers functionality to manage a single child process. The worker is
    responsible for communicating heartbeat information with it's child
    process and starting, stopping and restarting a child as required or
    instructed.
    """

    _started = False
    ping_count = 0

    def __init__(self, idx, socks, loop=None):
        """
        Initialise the worker.

        :param str idx: The number reference of the worker and child.
        :param list socks: Sockets to listen for connections on.
        :param loop: The event loop to use.
        :type loop: :py:class:`asyncio.unix_events._UnixSelectorEventLoop` or
                    :py:obj:`None` to get the current event loop using
                    :py:func:`asyncio.get_event_loop`.
        """
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self.socks = socks
        self.idx = idx
        self.start()

    def start(self):
        """Create and fork off a child process for the current worker."""
        if self._started is True:
            raise AssertionError("Cannot start an already started worker")
        self._started = True

        self.up_read, self.up_write = os.pipe()
        self.down_read, self.down_write = os.pipe()

        self.pid = os.fork()
        if self.pid > 0:  # Parent
            asyncio.ensure_future(self.connect())
        else:  # Child
            self.setup_child()

    def setup_child(self):
        """Set the gid, uid and start the child process.."""
        setgid()
        setuid()
        asyncio.set_event_loop(None)
        if setproctitle:
            setproctitle.setproctitle("blackhole: worker")
        process = Child(self.up_read, self.down_write, self.socks, self.idx)
        process.start()

    def restart_child(self):
        """Restart the child process."""
        self.kill_child()
        self.pid = os.fork()
        if self.pid > 0:
            self.ping_count = 0
            self.ping = time.monotonic()
        else:
            self.setup_child()

    def kill_child(self):
        """Kill the child process."""
        try:
            os.kill(self.pid, signal.SIGTERM)
            os.wait()
        except ProcessLookupError:
            pass

    async def heartbeat(self, writer):
        """
        Handle heartbeat between a worker and child.

        If a child process stops communicating with it's worker, it will be
        killed, the worker managing it will also be removed and a new worker
        and child will be spawned.

        :param asyncio.StreamWriter writer: An object for writing data to the
                                            pipe.

        .. note::

           3 bytes are used in the communication channel.

           - b'x01' -- :const:`blackhole.protocols.PING`
           - b'x02' -- :const:`blackhole.protocols.PONG`

           The worker will sleep for 15 seconds, before requesting a ping from
           the child. If we go for over 30 seconds waiting for a ping, the
           worker will restart itself and the child bound to it.

           These message values are defined in the :mod:`blackhole.protocols`
           schema. Documentation is available at --
           https://kura.gg/blackhole/api-protocols.html
        """
        while self._started:
            await asyncio.sleep(15)
            if (time.monotonic() - self.ping) < 30:
                writer.write(protocols.PING)
            else:
                if self._started:
                    logger.debug(
                        f"worker.{self.idx}.heartbeat: Communication failed. "
                        "Restarting worker",
                    )
                    self.restart_child()

    async def chat(self, reader):
        """
        Communicate between a worker and child.

        If communication with the child fails the worker is shutdown and the
        child is killed.

        :param asyncio.StreamReader reader: An object for reading data from
                                            the pipe.

        .. note::

           3 bytes are used in the communication channel.

           - b'x01' -- :const:`blackhole.protocols.PING`
           - b'x02' -- :const:`blackhole.protocols.PONG`

           Read data coming in from the child. If a PONG is received, we'll
           update the worker, setting this PONG as a 'PING' from the child.

           These message values are defined in the :mod:`blackhole.protocols`
           schema. Documentation is available at --
           https://kura.gg/blackhole/api-protocols.html
        """
        while self._started:
            try:
                msg = await reader.read(3)
                if msg == protocols.PONG:
                    logger.debug(
                        f"worker.{self.idx}.chat: Pong received from child",
                    )
                    self.ping = time.monotonic()
                    self.ping_count += 1
            except:  # noqa
                self.stop()
            await asyncio.sleep(5)

    async def connect(self):
        """
        Connect the child and worker so they can communicate.

        :param int up_write: A file descriptor.
        :param int down_read: A file descriptor.
        """
        read_fd = os.fdopen(self.down_read, "rb")
        r_trans, r_proto = await self.loop.connect_read_pipe(
            StreamProtocol,
            read_fd,
        )
        write_fd = os.fdopen(self.up_write, "wb")
        w_trans, w_proto = await self.loop.connect_write_pipe(
            StreamProtocol,
            write_fd,
        )
        reader = r_proto.reader
        writer = asyncio.StreamWriter(w_trans, w_proto, reader, self.loop)
        self.ping = time.monotonic()
        self.rtransport = r_trans
        self.wtransport = w_trans
        self.chat_task = asyncio.ensure_future(self.chat(reader))
        self.heartbeat_task = asyncio.ensure_future(self.heartbeat(writer))

    def stop(self):
        """Terminate the worker and it's respective child process."""
        self.kill_child()
        self._started = False
        self.chat_task.cancel()
        self.heartbeat_task.cancel()
        self.rtransport.close()
        self.wtransport.close()



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Blackhole is an email MTA that pipes all mail to /dev/null.

Blackhole is built on top of asyncio and utilises :py:obj:`async def` and
:py:obj:`await` statements on available in Python 3.5 and above.

While Blackhole is an MTA (mail transport agent), none of the actions
performed of SMTP or SMTPS are actually processed and no email or sent or
delivered.
"""

from .application import __all__ as __application_all__
from .child import __all__ as __child_all__
from .config import __all__ as __config_all__
from .control import __all__ as __control_all__
from .daemon import __all__ as __daemon_all__
from .exceptions import __all__ as __exceptions_all__
from .logs import __all__ as __logs_all__
from .protocols import __all__ as __protocols_all__
from .smtp import __all__ as __smtp_all__
from .streams import __all__ as __streams_all__
from .supervisor import __all__ as __supervisor_all__
from .utils import __all__ as __utils_all__
from .worker import __all__ as __worker_all__


__author__ = "Kura"
__copyright__ = "None"
__credits__ = ("Kura",)
__license__ = "MIT"
__version__ = "2.1.19"
__maintainer__ = "Kura"
__email__ = "kura@kura.gg"
__status__ = "Stable"


__all__ = (
    __application_all__
    + __child_all__
    + __config_all__
    + __control_all__
    + __daemon_all__
    + __exceptions_all__
    + __logs_all__
    + __protocols_all__
    + __smtp_all__
    + __streams_all__
    + __supervisor_all__
    + __utils_all__
    + __worker_all__
)
"""Tuple all the things."""

try:
    import asyncio

    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['blackhole']

package_data = \
{'': ['*']}

extras_require = \
{'dev': ['uvloop>=0.14.0,<0.15.0',
         'setproctitle>=1.1.10,<2.0.0',
         'coverage>=5.1,<6.0',
         'pytest>=5.4.2,<6.0.0',
         'pytest-asyncio>=0.12.0,<0.13.0',
         'pytest-cov>=2.8.1,<3.0.0',
         'pytest-clarity>=0.3.0-alpha.0,<0.4.0',
         'sphinx>=3.0.3,<4.0.0',
         'guzzle_sphinx_theme>=0.7.11,<0.8.0',
         'tox>=3.15.0,<4.0.0',
         'pre-commit>=2.4.0,<3.0.0',
         'interrogate>=1.3.0,<2.0.0',
         'pyroma>=2.6,<3.0',
         'bandit>=1.6.2,<2.0.0',
         'black>=20.8b1,<21.0',
         'flake8>=3.8.3,<4.0.0',
         'flake8-bugbear>=20.1.4,<21.0.0',
         'flake8-isort>=4.0.0,<5.0.0',
         'flake8-commas>=2.0.0,<3.0.0',
         'pydocstyle>=5.1.0,<6.0.0',
         'doc8>=0.8.1,<0.9.0',
         'codespell>=2.1.0,<3.0.0',
         'vulture>=2.3,<3.0'],
 'docs': ['sphinx>=3.0.3,<4.0.0', 'guzzle_sphinx_theme>=0.7.11,<0.8.0'],
 'setproctitle': ['setproctitle>=1.1.10,<2.0.0'],
 'tests': ['coverage>=5.1,<6.0',
           'pytest>=5.4.2,<6.0.0',
           'pytest-asyncio>=0.12.0,<0.13.0',
           'pytest-cov>=2.8.1,<3.0.0',
           'pytest-clarity>=0.3.0-alpha.0,<0.4.0',
           'interrogate>=1.3.0,<2.0.0',
           'pyroma>=2.6,<3.0',
           'bandit>=1.6.2,<2.0.0',
           'black>=20.8b1,<21.0',
           'flake8>=3.8.3,<4.0.0',
           'flake8-bugbear>=20.1.4,<21.0.0',
           'flake8-isort>=4.0.0,<5.0.0',
           'flake8-commas>=2.0.0,<3.0.0',
           'pydocstyle>=5.1.0,<6.0.0',
           'doc8>=0.8.1,<0.9.0',
           'codespell>=2.1.0,<3.0.0',
           'vulture>=2.3,<3.0'],
 'uvloop': ['uvloop>=0.14.0,<0.15.0']}

entry_points = \
{'console_scripts': ['blackhole = blackhole.application:run',
                     'blackhole_config = '
                     'blackhole.application:blackhole_config']}

setup_kwargs = {
    'name': 'blackhole',
    'version': '2.1.19',
    'description': 'Blackhole is an MTA (message transfer agent) that (figuratively) pipes all mail to /dev/null.',
    'long_description': "=========\nBlackhole\n=========\n\n.. image:: https://img.shields.io/github/workflow/status/kura/blackhole/CI?style=for-the-badge&label=tests&logo=githubactions\n    :target: https://github.com/kura/blackhole/actions/workflows/ci.yml\n    :alt: Build status of the master branch\n\n.. image:: https://img.shields.io/codecov/c/github/kura/blackhole/master.svg?style=for-the-badge&label=coverage&logo=codecov\n    :target: https://codecov.io/github/kura/blackhole/\n    :alt: Test coverage\n\nBlackhole is an `MTA (message transfer agent)\n<https://en.wikipedia.org/wiki/Message_transfer_agent>`_ that (figuratively)\npipes all mail to /dev/null, built on top of `asyncio\n<https://docs.python.org/3/library/asyncio.html>`_ and utilises `async def <https://docs.python.org/3/reference/compound_stmts.html#async-def>`_\nand `await\n<https://docs.python.org/3/reference/expressions.html#await>`_ statements\navailable in `Python 3.5 <https://docs.python.org/3/whatsnew/3.5.html>`_.\n\nWhile Blackhole is an MTA, none of the actions performed via SMTP or SMTPS are\nactually processed, and no email is delivered. You can tell Blackhole how to\nhandle mail that it receives. It can accept all of it, bounce it all, or\nrandomly do either of those two actions.\n\nThink of Blackhole sort of like a honeypot in terms of how it handles mail, but\nit's specifically designed with testing in mind.\n\nPython support\n==============\n\n- Python 3.7+\n- PyPy 3.7+\n- Pyston 2.2+\n\nDocumentation\n=============\n\nYou can find the latest documentation `here\n<https://kura.gg/blackhole/>`_.\n\nIf you would like to contribute, please read the `contributors guide\n<https://kura.gg/blackhole/overview.html#contributing>`_.\n\nThe latest build status on GitHub `<https://github.com/kura/blackhole/actions/workflows/ci.yml>`_.\n\nAnd the test coverage report on `codecov\n<https://codecov.io/github/kura/blackhole/>`_.\n\nChangelog\n=========\n\nYou can find a list of changes `on the\nblackhole website <https://kura.gg/blackhole/changelog.html>`_.\n",
    'author': 'Kura',
    'author_email': 'kura@kura.gg',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://kura.gg/blackhole/',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>3.7,<4',
}


setup(**setup_kwargs)
